from django.utils.translation import pgettext_lazy
from wcd_notifications.compat import JSONField

from .. import channels
from ..utils import DynamicChoicesField, get_json_encoder
from .utils import DateTimeModel


__all__ = 'ChannelConfig',


class ChannelConfig(DateTimeModel):
    class Meta:
        verbose_name = pgettext_lazy('wcd_envoyer', 'Channel config')
        verbose_name_plural = pgettext_lazy('wcd_envoyer', 'Channel configs')

    channel = DynamicChoicesField(
        pgettext_lazy('wcd_envoyer', 'Channel'),
        max_length=512, unique=True,
        choices=channels.registry.get_choices,
    )
    data = JSONField(
        pgettext_lazy('wcd_envoyer', 'Data'), default=dict,
        blank=True, null=False, encoder=get_json_encoder(),
    )

    def __str__(self) -> str:
        return self.get_channel_display()



from django.apps import AppConfig
from django.utils.translation import pgettext_lazy


__all__ = 'ActionsTrackerConfig',


class ActionsTrackerConfig(AppConfig):
    name = 'wcd_envoyer.contrib.pxd_actions_tracker'
    label = 'wcd_envoyer_actions_tracker'
    verbose_name = pgettext_lazy('wcd_envoyer', 'Envoyer actions tracker')

    def ready(self) -> None:
        from . import handlers



from typing import *
import json

from px_domains import Domain
from wcd_envoyer.services.sender import FailedMessagesGroup
from wcd_envoyer.models import Message
from wcd_envoyer.utils import get_json_encoder
from pxd_actions_tracker.cases import log


MESSAGE_FORMAT = 'Failed send "{message[event]}" through "{message[channel]}"'.format


def format_error(error: Union[Exception, dict, str]):
    if isinstance(error, Exception):
        return {'exception': str(error)}

    # FIXME!
    return json.loads(json.dumps(error, cls=get_json_encoder()))


def resolve_log_detail(message: Message, error):
    return {
        'message': {
            'channel': message.channel,
            'event': message.event,
            'recipients': message.recipients,
            'data': message.data,
        },
        'error': format_error(error),
    }


def track_failed_messages_groups(
    domain: Domain,
    groups: Sequence[FailedMessagesGroup],
    message_format: Callable = MESSAGE_FORMAT,
    detail_resolver: Callable = resolve_log_detail,
):
    for backend, config, messages in groups:
        for message, error in messages:
            detail = detail_resolver(message, error)
            log(domain, message=message_format(**detail), detail=detail)



# Generated by Django 2.2 on 2023-05-03 13:14

import django.contrib.postgres.fields.jsonb
from django.db import migrations, models
import wcd_envoyer.const
import wcd_envoyer.utils.models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='ChannelConfig',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created_at', models.DateTimeField(auto_now_add=True, verbose_name='Created at')),
                ('updated_at', models.DateTimeField(auto_now=True, verbose_name='Updated at')),
                ('channel', wcd_envoyer.utils.models.DynamicChoicesField(choices=[], max_length=512, unique=True, verbose_name='Channel')),
                ('data', django.contrib.postgres.fields.jsonb.JSONField(default=dict, verbose_name='Data')),
            ],
            options={
                'verbose_name': 'Channel config',
                'verbose_name_plural': 'Channel configs',
            },
        ),
        migrations.CreateModel(
            name='Message',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created_at', models.DateTimeField(auto_now_add=True, verbose_name='Created at')),
                ('updated_at', models.DateTimeField(auto_now=True, verbose_name='Updated at')),
                ('status', models.CharField(choices=[('010-new', 'New'), ('030-pending', 'Pending'), ('060-sent', 'Sent'), ('080-failed', 'Failed')], default=wcd_envoyer.const.MessageStatus('010-new'), max_length=512, verbose_name='Status')),
                ('channel', wcd_envoyer.utils.models.DynamicChoicesField(choices=[], max_length=512, verbose_name='Channel')),
                ('event', wcd_envoyer.utils.models.DynamicChoicesField(choices=[], max_length=512, verbose_name='Event')),
                ('recipients', django.contrib.postgres.fields.jsonb.JSONField(default=list, verbose_name='Recipients')),
                ('data', django.contrib.postgres.fields.jsonb.JSONField(default=dict, verbose_name='Data')),
            ],
            options={
                'verbose_name': 'Message',
                'verbose_name_plural': 'Messages',
            },
        ),
        migrations.CreateModel(
            name='Template',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created_at', models.DateTimeField(auto_now_add=True, verbose_name='Created at')),
                ('updated_at', models.DateTimeField(auto_now=True, verbose_name='Updated at')),
                ('channel', wcd_envoyer.utils.models.DynamicChoicesField(choices=[], max_length=512, verbose_name='Channel')),
                ('event', wcd_envoyer.utils.models.DynamicChoicesField(choices=[], max_length=512, verbose_name='Event')),
                ('data', django.contrib.postgres.fields.jsonb.JSONField(default=dict, verbose_name='Data')),
                ('is_active', models.BooleanField(default=False, verbose_name='Is active')),
            ],
            options={
                'verbose_name': 'Template',
                'verbose_name_plural': 'Templates',
                'unique_together': {('channel', 'event')},
            },
        ),
    ]



from dataclasses import dataclass, field
from typing import *

from px_settings.contrib.django import settings as setting_wrap

from .channels import ChannelInfoType
from .events import EventInfoType
from .const import SETTINGS_PREFIX


__all__ = 'Settings', 'settings',


@setting_wrap(SETTINGS_PREFIX)
@dataclass
class Settings:
    CHANNELS: Sequence[ChannelInfoType] = field(default_factory=list)
    EVENTS: Sequence[EventInfoType] = field(default_factory=list)
    JSON_ENCODER: str = 'wcd_envoyer.utils.types.EnvoyerJSONEncoder'


settings = Settings()



import pytest

from celery import Celery

from wcd_envoyer.contrib.celery.shortcuts import task, make_task
from wcd_envoyer.models import ChannelConfig, Message

from .const import CONSOLE_CHANNEL
from .utils import make_template


EVENT_ONE = 'unexisting_event'


@pytest.fixture
@pytest.mark.django_db
def simple_console_config():
    config = ChannelConfig.objects.create(channel=CONSOLE_CHANNEL)

    return *make_template(CONSOLE_CHANNEL, EVENT_ONE), config


@pytest.mark.django_db
def test_celery_send(django_assert_num_queries, simple_console_config, mocker):
    app = Celery('app')
    app.conf.update(
        task_always_eager=True,
        broker_url='memory://',
        result_backend='cache+memory://',
    )
    runner_task = make_task(app)
    send = runner_task.send

    with django_assert_num_queries(9):
        mocker.patch(
            'wcd_envoyer.channels.backends.console.ConsoleBackend.send',
            lambda self, config, messages, context: (messages, []),
        )
        send(
            EVENT_ONE,
            [{'user_id': 1}, {'user_id': 2}],
            {'username': 'Named', 'language': 'ja'},
        )

    assert Message.objects.all().count() == 2
    assert Message.objects.filter(status=Message.Status.SENT).count() == 2


@pytest.mark.django_db
def test_celery_callback(django_assert_num_queries, simple_console_config, mocker):
    app = Celery('app')
    app.conf.update(
        task_always_eager=True,
        broker_url='memory://',
        result_backend='cache+memory://',
    )
    saved = {}

    @task(app)
    def runner_task(result, context):
        saved['succeeded'], saved['failed'] = result

    send = runner_task.send

    with django_assert_num_queries(9):
        mocker.patch(
            'wcd_envoyer.channels.backends.console.ConsoleBackend.send',
            lambda self, config, messages, context: (messages, []),
        )
        send(
            EVENT_ONE,
            [{'user_id': 1}, {'user_id': 2}],
            {'username': 'Named', 'language': 'ja'},
        )

    assert Message.objects.all().count() == 2
    assert Message.objects.filter(status=Message.Status.SENT).count() == 2
    assert len(saved) == 2



from typing import *

from .services.sender import Sender
from .channels import MessageData, Recipients, BaseMessagingBackend


__all__ = 'default_sender', 'send_multiple', 'send', 'get_backend',


default_sender = Sender()
send_multiple = default_sender.send


def send(
    event: str,
    recipients: Recipients,
    context: dict,
    channels: Optional[Sequence[str]] = None,
    sender: Sender = default_sender,
):
    all_channels = (
        channels
        if channels is not None else
        sender.get_backends().keys()
    )

    return sender.send(
        [
            MessageData(channel, event, recipients, context)
            for channel in all_channels
        ],
        context=context,
        only_channels=channels,
    )


def get_backend(
    channel: str,
    sender: Sender = default_sender,
) -> Optional[BaseMessagingBackend]:
    return sender.get_backends(only_channels=[channel]).get(channel, None)



from typing import *
from pprint import pprint

from wcd_envoyer.models import Message, ChannelConfig

from ..backend import BaseMessagingBackend


def print_heading(message, l=80, s='='):
    print((message + ' ' + (s*l)).strip()[:l])


class ConsoleBackend(BaseMessagingBackend):
    def send(self, config: ChannelConfig, messages: Sequence[Message], context={}):
        print_heading('Printing %s messages' % len(messages))

        for message in messages:
            print_heading('', s='-')
            print('ID: %s' % message.pk)
            print('Channel: %s' % message.channel)
            print('Event: %s' % message.event)
            print('Status: %s' % message.status)
            print('To:')
            pprint(message.recipients, compact=True, indent=2)
            print('Data:')
            pprint(message.data, compact=True, indent=2)
            print_heading('', s='-')

        print_heading('')

        return messages, []



from django.apps import AppConfig
from django.utils.translation import pgettext_lazy


__all__ = 'EnvoyerConfig',


class EnvoyerConfig(AppConfig):
    name = 'wcd_envoyer'
    verbose_name = pgettext_lazy('wcd_envoyer', 'Envoyer')

    def ready(self) -> None:
        super().ready()

        self.module.autodiscover()



from typing import *
from functools import partialmethod

from django.db import models


__all__ = (
    'DynamicChoicesField',
)


class DynamicChoicesField(models.CharField):
    _choices_resolver: Callable

    @property
    def choices(self):
        result = self._choices_resolver()
        return result if isinstance(result, list) else list(result)

    @choices.setter
    def choices(self, value: Callable):
        if isinstance(value, (list, tuple)):
            saved = value
            def value(): return saved

        self._choices_resolver = value

    def deconstruct(self):
        name, path, args, kwargs = super().deconstruct()
        kwargs["choices"] = []
        return name, path, args, kwargs

    def contribute_to_class(self, cls, name, private_only=False):
        super().contribute_to_class(cls, name, private_only=private_only)

        mname = 'get_%s_display' % self.name

        if not hasattr(cls, mname):
            setattr(cls, mname, partialmethod(cls._get_FIELD_display, field=self))



from django.db import models
from django.utils.translation import pgettext_lazy


__all__ = 'DateTimeModel',


class DateTimeModel(models.Model):
    class Meta:
        abstract = True

    created_at = models.DateTimeField(
        pgettext_lazy('wcd_envoyer', 'Created at'), auto_now_add=True,
    )
    updated_at = models.DateTimeField(
        pgettext_lazy('wcd_envoyer', 'Updated at'), auto_now=True,
    )



from typing import *
from functools import reduce
from operator import or_

from django.db import models

from wcd_envoyer.models import Template
from wcd_envoyer.utils import KwargsInjector


ClientEventPair = Tuple[str, str]
Templates = Dict[ClientEventPair, Template]


class TemplatesResolver(KwargsInjector):
    queryset = (
        Template.objects
        .filter(is_active=True)
        .prefetch_related('translated')
    )

    def get_templates(self, pairs: Iterable[ClientEventPair]) -> Templates:
        items = self.queryset.filter(reduce(or_, (
            models.Q(channel=channel, event=event)
            for channel, event in pairs
        )))

        return {
            (template.channel, template.event): template
            for template in items
        }



from typing import *
from functools import partial, wraps

from wcd_envoyer.shortcuts import send

from .services.sender import CelerySender
from .tasks import actual_sending_messages_task

if TYPE_CHECKING:
    from celery import Celery


__all__ = 'make_task', 'task',


def make_task(
    app: 'Celery',
    apply_kwargs: dict = {},
    task_kwargs: dict = {},
    task_callback: Optional[Callable] = None,
    **kwargs,
):
    sender = CelerySender(
        celery_task_caller=lambda *a, **kw: task.apply_async(
            args=a, kwargs=kw, **apply_kwargs
        ),
        **kwargs,
    )
    to_wrap = task_callback or actual_sending_messages_task
    fn = wraps(to_wrap)(lambda *a, **kw: actual_sending_messages_task(
        *a, **kw, sender=sender, callback=task_callback,
    ))

    task = app.task(**task_kwargs)(fn)
    task.sender = sender
    task.send = partial(send, sender=sender)

    return task


def task(
    app: 'Celery',
    apply_kwargs: dict = {},
    task_kwargs: dict = {},
    **kwargs,
):
    return lambda task_callback: make_task(
        app, apply_kwargs=apply_kwargs, task_kwargs=task_kwargs, **kwargs,
        task_callback=task_callback,
    )



import pytest

from django.test import override_settings
from wcd_envoyer.models import ChannelConfig, Message
from wcd_envoyer.shortcuts import send
from django.core import mail

from .const import CONSOLE_CHANNEL, EMAIL_CHANNEL
from .utils import make_template


EVENT_ONE = 'unexisting_event'


@pytest.fixture
@pytest.mark.django_db
def simple_console_config():
    config = ChannelConfig.objects.create(channel=CONSOLE_CHANNEL)

    return *make_template(CONSOLE_CHANNEL, EVENT_ONE), config


@pytest.fixture
@pytest.mark.django_db
def simple_email_config():
    config = ChannelConfig.objects.create(channel=EMAIL_CHANNEL, data={
        'from_email': 'galina@blanka.bulbul',
        'reply_to': ['John Doe <john@doe.com>', 'bubel@gmail.com'],
    })

    return *make_template(EMAIL_CHANNEL, EVENT_ONE), config


@pytest.mark.django_db
def test_simple_sender(django_assert_num_queries, simple_console_config, mocker):
    with django_assert_num_queries(6):
        mocker.patch(
            'wcd_envoyer.channels.backends.console.ConsoleBackend.send',
            lambda self, config, messages, context: (messages, []),
        )
        send(
            EVENT_ONE,
            [{'user_id': 1}, {'user_id': 2}],
            {'username': 'Named', 'language': 'ja'},
        )

    assert Message.objects.all().count() == 2
    assert Message.objects.filter(status=Message.Status.SENT).count() == 2


@pytest.mark.django_db
def test_succeed_failed_lists(django_assert_num_queries, simple_console_config, mocker):
    with django_assert_num_queries(6):
        mocker.patch(
            'wcd_envoyer.channels.backends.console.ConsoleBackend.send',
            lambda self, config, messages, context: (messages[:1], [(messages[1], None)]),
        )
        send(
            EVENT_ONE,
            [{'user_id': 1}, {'user_id': 2}],
            {'username': 'Named', 'language': 'ja'},
        )

    assert Message.objects.all().count() == 2
    assert Message.objects.filter(status=Message.Status.SENT).count() == 1
    assert Message.objects.filter(status=Message.Status.FAILED).count() == 1


@pytest.mark.django_db
def test_send_exception(django_assert_num_queries, simple_console_config, mocker):
    with django_assert_num_queries(6):
        mocker.patch(
            'wcd_envoyer.channels.backends.console.ConsoleBackend.send',
            lambda self, config, messages, context: (_ for _ in ()).throw(Exception()),
        )
        send(
            EVENT_ONE,
            [{'user_id': 1}, {'user_id': 2}],
            {'username': 'Named', 'language': 'ja'},
        )

    assert Message.objects.all().count() == 2
    assert Message.objects.filter(status=Message.Status.FAILED).count() == 2


@pytest.mark.django_db
def test_inactive_template(django_assert_num_queries, simple_console_config):
    template, _, _ = simple_console_config
    template.is_active = False
    template.save()

    with django_assert_num_queries(2):
        send(
            EVENT_ONE,
            [{'user_id': 1}, {'user_id': 2}],
            {'username': 'Named'},
        )

    assert Message.objects.all().count() == 0


@pytest.mark.django_db
def test_email(django_assert_num_queries, simple_email_config):
    with django_assert_num_queries(6):
        mail.outbox.clear()

        with override_settings(
            EMAIL_BACKEND='django.core.mail.backends.locmem.EmailBackend'
        ):
            send(
                EVENT_ONE,
                [{'email': 'ad@ad.ad'}, {'user_id': 2}],
                {'username': 'Named', 'language': 'ja'},
            )

    assert Message.objects.all().count() == 1
    assert Message.objects.filter(status=Message.Status.SENT).count() == 1
    assert len(mail.outbox) == 1

    msg = str(mail.outbox[0].message())

    assert '[ja] NOt a plain. [ja] Hi Named test_email !' in msg
    assert 'From: galina@blanka.bulbul' in msg
    assert 'To: ad@ad.ad' in msg
    assert 'Reply-To: John Doe <john@doe.com>, bubel@gmail.com' in msg



from django import forms

from wcd_envoyer.models import Template, ChannelConfig


__all__ = 'AddTemplateForm', 'AddConfigForm',


class AddTemplateForm(forms.ModelForm):
    class Meta:
        fields = 'channel', 'event', 'is_active',
        model = Template


class AddConfigForm(forms.ModelForm):
    class Meta:
        fields = 'channel',
        model = ChannelConfig



__version__ = '0.2.5'

from django.utils.module_loading import autodiscover_modules

VERSION = tuple(__version__.split('.'))

default_app_config = 'wcd_envoyer.apps.EnvoyerConfig'


def autodiscover():
    from .conf import settings
    from . import channels, events

    autodiscover_modules('envoyer')

    for event in settings.EVENTS:
        events.registry.add(event)

    for channel in settings.CHANNELS:
        channels.registry.add(channel)



# Generated by Django 2.2 on 2023-05-05 05:52

import django.contrib.postgres.fields.jsonb
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('wcd_envoyer', '0001_initial'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='message',
            options={'ordering': ('-created_at',), 'verbose_name': 'Message', 'verbose_name_plural': 'Messages'},
        ),
        migrations.CreateModel(
            name='TemplateTranslation',
            fields=[
                ('id', models.BigAutoField(primary_key=True, serialize=False, verbose_name='ID')),
                ('language', models.CharField(choices=[('ru', 'Russian'), ('uk', 'Ukrainian'), ('pl', 'Polish'), ('cs', 'Czech'), ('ru-md', 'Russian-md'), ('ru-uz', 'Russian-uz'), ('cs-sk', 'Czech-sk'), ('ro', 'Romanian')], default='ru', max_length=32, verbose_name='Language')),
                ('data', django.contrib.postgres.fields.jsonb.JSONField(default=dict, verbose_name='Data')),
                ('entity', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='translations', to='wcd_envoyer.Template', verbose_name='Translation entity')),
            ],
            options={
                'base_manager_name': 'objects',
                'unique_together': {('language', 'entity')},
            },
        ),
    ]
